
from axelrod import Player, random_choice
import random



class AverageCopier(Player):
    """The player will cooperate with probability p if the opponent's cooperation ratio is p."""

    name = 'Average Copier'
    memory_depth = float('inf')  # Long memory

    @staticmethod
    def strategy(opponent):
        """Randomly picks a strategy (not affected by history)."""
        if len(opponent.history) == 0:
            return random_choice(0.5)
        p = opponent.cooperations // len(opponent.history)
        rnd_num = random.random()
        if rnd_num < p:
            return 'C'
        return 'D'

class NiceAverageCopier(Player):
    """Same as Average Copier, but always starts by cooperating."""

    name = 'Nice Average Copier'
    memory_depth = float('inf')  # Long memory

    @staticmethod
    def strategy(opponent):
        """Randomly picks a strategy (not affected by history)."""
        if len(opponent.history) == 0:
            return 'C'
        p = opponent.cooperations // len(opponent.history)
        return random_choice(p)



from axelrod import Player


class GoByMajority(Player):
    """A player examines the history of the opponent: if the opponent has more
    defections than cooperations then the player defects.

    An optional memory attribute will limit the number of turns remembered (by
    default this is 0)
    """

    # memory_depth is set by __init__

    def __init__(self, memory_depth=0, soft=True):
        Player.__init__(self)
        self.soft = soft
        self.memory_depth = memory_depth
        # Set class var for consistency with other strategies
        self.__class__.memory_depth = memory_depth

    def strategy(self, opponent):
        """This is affected by the history of the opponent.

        As long as the opponent cooperates at least as often as they defect then the player will cooperate.
        If at any point the opponent has more defections than cooperations in memory the player defects.
        """

        memory = self.memory_depth
        history = opponent.history[-memory:]
        defections = sum([s == 'D' for s in history])
        cooperations = sum([s == 'C' for s in history])
        if defections > cooperations:
            return 'D'
        if defections == cooperations:
            if self.soft:
                return 'C'
            else:
                return 'D'
        return 'C'

    def __repr__(self):
        """The string method for the strategy."""
        memory = self.memory_depth
        return 'Go By Majority' + (memory > 0) * (":%i" % memory)


class GoByMajority40(GoByMajority):
    """ 
    GoByMajority player with a memory of 40.
    """

    def __init__(self, memory_depth=40):
        super(self.__class__, self).__init__(memory_depth=memory_depth)


class GoByMajority20(GoByMajority):
    """
    GoByMajority player with a memory of 20.
    """

    def __init__(self, memory_depth=20):
        super(self.__class__, self).__init__(memory_depth=memory_depth)


class GoByMajority10(GoByMajority):
    """
    GoByMajority player with a memory of 10.
    """

    def __init__(self, memory_depth=10):
        super(self.__class__, self).__init__(memory_depth=memory_depth)


class GoByMajority5(GoByMajority):
    """
    GoByMajority player with a memory of 5.
    """

    def __init__(self, memory_depth=5):
        super(self.__class__, self).__init__(memory_depth=memory_depth)



import math
import csv

try:
    # Python 2
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO


def median(lst):
    lst = sorted(lst)
    if len(lst) < 1:
            return None
    if len(lst) % 2 == 1:
            return lst[((len(lst)+1) // 2)-1]
    if len(lst) % 2 == 0:
            return float(sum(lst[(len(lst) // 2)-1:(len(lst) // 2)+1]))/2.0


class ResultSet(object):
    """A class to hold the results of a tournament."""

    def __init__(self, players, turns, repetitions, outcome):
        self.players = players
        self.nplayers = len(players)
        self.turns = turns
        self.repetitions = repetitions
        self.outcome = outcome
        self.results = self._results(outcome)
        if 'payoff' in self.results:
            self.scores = self._scores(self.results['payoff'])
            self.normalised_scores = self._normalised_scores(self.scores)
            self.ranking = self._ranking(self.scores)
            self.ranked_names = self._ranked_names(self.ranking)
            self.payoff_matrix, self.payoff_stddevs = (
                self._payoff_matrix(self.results['payoff']))
        if 'cooperation' in self.results:
            self.cooperation = self._cooperation(self.results['cooperation'])
            self.normalised_cooperation = (
                self._normalised_cooperation(self.cooperation))

    def _null_matrix(self):
        plist = list(range(self.nplayers))
        replist = list(range(self.repetitions))
        return [[[0 for r in replist] for j in plist] for i in plist]

    def _results(self, outcome):
        results = {}
        for result_type, result_list in outcome.items():
            matrix = self._null_matrix()
            for index, result_matrix in enumerate(result_list):
                for i in range(len(self.players)):
                    for j in range(len(self.players)):
                        matrix[i][j][index] = result_matrix[i][j]
                results[result_type] = matrix
        return results

    def _scores(self, payoff):
        """Return scores based on the results.

        Originally there were no self-interactions, so the code here was
        rewritten to exclude those from the generated score. To include
        self-interactions, remove the condition on ip and ires and fix the
        normalization factor.
        """
        scores = []
        for ires, res in enumerate(payoff):
            scores.append([])
            for irep in range(self.repetitions):
                scores[-1].append(0)
                for ip in range(self.nplayers):
                    if ip != ires:
                        scores[-1][-1] += res[ip][irep]
        return scores

    def _normalised_scores(self, scores):
        normalisation = self.turns * (self.nplayers - 1)
        return [
            [1.0 * s / normalisation for s in r] for r in scores]

    def _ranking(self, scores):
        """
        Returns a list of players (their index within the
        players list rather than a player instance)
        ordered by median score
        """
        ranking = sorted(
            range(self.nplayers),
            key=lambda i: -median(scores[i]))
        return ranking

    def _ranked_names(self, ranking):
        """Returns a list of players names sorted by their ranked order."""
        ranked_names = [str(self.players[i]) for i in ranking]
        return ranked_names

    def _payoff_matrix(self, payoff):
        """Returns a per-turn averaged payoff matrix and its stddevs."""
        averages = []
        stddevs = []
        for res in payoff:
            averages.append([])
            stddevs.append([])
            for s in res:
                perturn = [1.0 * rep / self.turns for rep in s]
                avg = sum(perturn) / self.repetitions
                dev = math.sqrt(
                    sum([(avg - pt)**2 for pt in perturn]) / self.repetitions)
                averages[-1].append(avg)
                stddevs[-1].append(dev)
        return averages, stddevs

    def _cooperation(self, results):
        return[[sum(element) for element in row] for row in results]

    def _normalised_cooperation(self, cooperation):
        normalisation = self.turns * self.repetitions
        return[
            [1.0 * element / normalisation for element in row]
            for row in cooperation]

    def csv(self):
        csv_string = StringIO()
        header = ",".join(self.ranked_names) + "\n"
        csv_string.write(header)
        writer = csv.writer(csv_string, lineterminator="\n")
        for irep in range(self.repetitions):
            data = [self.normalised_scores[rank][irep]
                    for rank in self.ranking]
            writer.writerow(list(map(str, data)))
        return csv_string.getvalue()



import itertools

from axelrod import Player
from .memoryone import Joss


class Calculator(Player):
    """
    Plays like (Hard) Joss for the first 20 rounds. If periodic behavior is
    detected, defect forever. Otherwise play TFT.
    """

    name = "Calculator"
    memory_depth = float('inf')

    def __init__(self):
        Player.__init__(self)
        self.joss_instance = Joss()

    @staticmethod
    def detect_cycle(history):
        """Detects if there is a cycle in the opponent's history."""
        for i in range(len(history) // 2):
            cycle = itertools.cycle(history[0: i + 1])
            cycle_list = list(itertools.islice(cycle, 0, len(history)))
            if list(history) == cycle_list:
                return True
        return False

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 20:
            self.cycle = self.detect_cycle(opponent.history)
            return self.extended_strategy(opponent)
        if turn > 20:
            return self.extended_strategy(opponent)
        else:
            play = self.joss_instance.strategy(opponent)
            self.joss_instance.history.append(play)
            return play

    def extended_strategy(self, opponent):
        if self.cycle:
            return 'D'
        else:
            # TFT
            return 'D' if opponent.history[-1:] == ['D'] else 'C'

    def reset(self):
        Player.reset(self)
        self.joss_instance = Joss()



from axelrod import Player
import random


class Inverse(Player):
    """A player who defects with a probability that diminishes relative to how
    long ago the opponent defected."""

    name = 'Inverse'
    memory_depth = float('inf')  # Long memory

    @staticmethod
    def strategy(opponent):
        """Looks at opponent history to see if they have defected.

        If so, player defection is inversely proportional to when this occurred.
        """

        index = next((index for index, value in enumerate(opponent.history, start=1) if value == 'D'), None)

        if index is None:
            return 'C'

        rnd_num = random.random()

        if rnd_num < 1 / float(abs(index)):
            return 'D'
        return 'C'



"""
Additional strategies from Axelrod's two tournaments.
"""

from axelrod import Player

import random


flip_dict = {'C': 'D', 'D': 'C'}

## First Tournament

class Davis(Player):
    """A player starts by cooperating for 10 rounds then plays Grudger,
    defecting if at any point the opponent has defected."""

    name = 'Davis'
    memory_depth = float('inf')  # Long memory

    def __init__(self, rounds_to_cooperate=10):
        Player.__init__(self)
        self._rounds_to_cooperate = rounds_to_cooperate

    def strategy(self, opponent):
        """Begins by playing C, then plays D for the remaining rounds if the
        opponent ever plays D."""
        if len(self.history) < self._rounds_to_cooperate:
            return 'C'
        if opponent.defections:
            return 'D'
        return 'C'


class Feld(Player):
    """
    Defects when opponent defects. Cooperates with a probability that decreases
    to 0.5 at round 200.
    """

    name = "Feld"
    memory_depth = 200 # Varies actually, eventually becomes depth 1

    def __init__(self, start_coop_prob=1.0, end_coop_prob=0.5,
                 rounds_of_decay=200):
        Player.__init__(self)
        self._start_coop_prob = start_coop_prob
        self._end_coop_prob = end_coop_prob
        self._rounds_of_decay = rounds_of_decay

    def _cooperation_probability(self):
        """It's not clear what the interpolating function is, so we'll do
        something simple that decreases monotonically from 1.0 to 0.5 over
        200 rounds."""
        diff = (self._end_coop_prob - self._start_coop_prob)
        slope = diff / float(self._rounds_of_decay)
        rounds = len(self.history)
        return max(self._start_coop_prob + slope * rounds,
                   self._end_coop_prob)

    def strategy(self, opponent):
        if not self.history:
            return 'C'
        if opponent.history[-1] == 'D':
            return 'D'
        p = self._cooperation_probability()
        r = random.random()
        if r < p:
            return 'C'
        return 'D'


class Shubik(Player):
    """
    Plays like Tit-For-Tat with the following modification. After
    each retaliation, the number of rounds that Shubik retaliates
    increases by 1.
    """

    name = 'Shubik'
    memory_depth = float('inf')

    def __init__(self):
        Player.__init__(self)
        self.is_retaliating = False
        self.retaliation_length = 0
        self.retaliation_remaining = 0

    def _decrease_retaliation_counter(self):
        """Lower the remaining owed retaliation count and flip to non-retaliate
        if the count drops to zero."""
        if self.is_retaliating:
            self.retaliation_remaining -= 1
            if self.retaliation_remaining == 0:
                self.is_retaliating = False

    def strategy(self, opponent):
        if not opponent.history:
            return 'C'
        if opponent.history[-1] == 'D':
            # Retaliate against defections
            if self.history[-1] == 'C': # it's on now!
                # Lengthen the retaliation period
                self.is_retaliating = True
                self.retaliation_length += 1
                self.retaliation_remaining = self.retaliation_length
                self._decrease_retaliation_counter()
                return 'D'
            else:
                # Just retaliate
                if self.is_retaliating:
                    self._decrease_retaliation_counter()
                return 'D'
        if self.is_retaliating:
            # Are we retaliating still?
            self._decrease_retaliation_counter()
            return 'D'
        return 'C'

    def reset(self):
        Player.reset(self)
        self.is_retaliating = False
        self.retaliation_length = 0
        self.retaliation_remaining = 0

class Tullock(Player):
    """
    Cooperates for the first 11 rounds then randomly cooperates 10% less often
    than the opponent has in previous rounds."""

    name = "Tullock"
    memory_depth = 11 # long memory, modified by init

    def __init__(self, rounds_to_cooperate=11):
        Player.__init__(self)
        self._rounds_to_cooperate = rounds_to_cooperate
        self.__class__.memory_depth = rounds_to_cooperate

    def strategy(self, opponent):
        rounds = self._rounds_to_cooperate
        if len(self.history) < rounds:
            return 'C'
        cooperate_count = opponent.history[-rounds:].count('C')
        prop_cooperate = cooperate_count / float(rounds)
        prob_cooperate = max(0, prop_cooperate - 0.10)
        r = random.random()
        if r < prob_cooperate:
            return 'C'
        return 'D'


## Second Tournament

class Champion(Player):
    """
    Strategy submitted to Axelrod's second tournament by Danny Champion.
    """

    name = "Champion"
    memory_depth = float('inf')

    def strategy(self, opponent):
        current_round = len(self.history)
        expected_length = self.tournament_length
        # Cooperate for the first 1/20-th of the game
        if current_round == 0:
            return 'C'
        if current_round < expected_length / 20.:
            return 'C'
        # Mirror partner for the next phase
        if current_round < expected_length * 5 / 40.:
            return opponent.history[-1]
        # Now cooperate unless all of the necessary conditions are true
        defection_prop = float(opponent.defections) / len(opponent.history)
        if opponent.history[-1] == 'D':
            r = random.random()
            if defection_prop > max(0.4, r):
                return 'D'
        return 'C'


class Eatherley(Player):
    """
    Strategy submitted to Axelrod's second tournament by Graham Eatherley.
    """

    name = "Eatherley"
    memory_depth = float('inf')

    def strategy(self, opponent):
        # Cooperate on the first move
        if not len(self.history):
            return 'C'
        # Reciprocate cooperation
        if opponent.history[-1] == 'C':
            return 'C'
        # Respond to defections with probability equal to opponent's total
        # proportion of defections
        defection_prop = float(opponent.defections) / len(opponent.history)
        r = random.random()
        if r < defection_prop:
            return 'D'
        return 'C'


class Tester(Player):
    """
    Submitted to Axelrod's second tournament by David Gladstein.

    Defects on the first move and plays TFT if the opponent ever defects (after
    one apology cooperation round). Otherwise alternate cooperation and defection.
    """

    name = "Tester"
    memory_depth = float('inf')

    def __init__(self):
        Player.__init__(self)
        self.is_TFT = False

    def strategy(self, opponent):
        # Defect on the first move
        if not opponent.history:
            return 'D'
        # Am I TFT?
        if self.is_TFT:
            return 'D' if opponent.history[-1:] == ['D'] else 'C'
        else:
            # Did opponent defect?
            if opponent.history[-1] == 'D':
                self.is_TFT = True
                return 'C'
            if len(self.history) in [1, 2]:
                return 'C'
            # Alternate C and D
            return flip_dict[self.history[-1]]

    def reset(self):
        Player.reset(self)
        self.is_TFT = False



from __future__ import absolute_import

from .alternator import *
from .appeaser import *
from .averagecopier import *
from .axelrod_tournaments import *
from .backstabber import *
from .calculator import Calculator
from .cooperator import *
from .darwin import *
from .defector import *
from .forgiver import *
from .geller import *
from .gobymajority import *
from .grudger import *
from .grumpy import *
from .hunter import *
from .inverse import *
from .mathematicalconstants import *
from .memoryone import *
from .meta import *
from .mindcontrol import *
from .mindreader import *
from .oncebitten import *
from .prober import *
from .punisher import *
from .qlearner import *
from .rand import *
from .retaliate import *
from .titfortat import *


basic_strategies = [
    Alternator,
    Cooperator,
    Defector,
    Random,
    TitForTat,
]

ordinary_strategies = [
    Aggravater,
    AlternatorHunter,
    Appeaser,
    AntiTitForTat,
    #ArrogantQLearner, # needs fixing
    AverageCopier,
    BackStabber,
    Bully,
    Calculator,
    #CautiousQLearner, # needs fixing
    Champion,
    CooperatorHunter,
    Davis,
    DefectorHunter,
    DoubleCrosser,
    Eatherley,
    FoolMeOnce,
    ForgetfulFoolMeOnce,
    ForgetfulGrudger,
    Forgiver,
    ForgivingTitForTat,
    GTFT,
    GoByMajority,
    GoByMajority10,
    GoByMajority20,
    GoByMajority40,
    GoByMajority5,
    Feld,
    Golden,
    Grofman,
    Grudger,
    Grumpy,
    HardProber,
    HardTitForTat,
    HardTitFor2Tats,
    #HesitantQLearner, # needs fixing
    Inverse,
    InversePunisher,
    Joss,
    LimitedRetaliate,
    LimitedRetaliate2,
    LimitedRetaliate3,
    MathConstantHunter,
    MetaHunter,
    MetaMajority,
    MetaMinority,
    MetaWinner,
    NiceAverageCopier,
    OnceBitten,
    OppositeGrudger,
    Pi,
    Prober,
    Prober2,
    Prober3,
    Punisher,
    RandomHunter,
    Retaliate,
    Retaliate2,
    Retaliate3,
    #RiskyQLearner, # Needs fixing
    Shubik,
    SoftJoss,
    SneakyTitForTat,
    StochasticWSLS,
    SuspiciousTitForTat,
    Tester,
    TitFor2Tats,
    TrickyCooperator,
    TrickyDefector,
    Tullock,
    TwoTitsForTat,
    WinStayLoseShift,
    ZDExtort2,
    ZDGTFT2,
    e,
]

# These are strategies that do not follow the rules of Axelrods tournament
cheating_strategies = [
    Darwin,
    Geller,
    GellerCooperator,
    GellerDefector,
    MindBender,
    MindController,
    MindReader,
    MindWarper,
    ProtectedMindReader,
]



from __future__ import absolute_import

from axelrod import Player
from .meta import MetaPlayer


class DefectorHunter(Player):
    """A player who hunts for defectors."""

    name = 'Defector Hunter'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        if len(self.history) >= 4 and len(opponent.history) == opponent.defections:
            return 'D'
        return 'C'


class CooperatorHunter(Player):
    """A player who hunts for cooperators."""

    name = 'Cooperator Hunter'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        if len(self.history) >= 4 and len(opponent.history) == opponent.cooperations:
            return 'D'
        return 'C'


class AlternatorHunter(Player):
    """A player who hunts for alternators."""

    name = 'Alternator Hunter'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        oh = opponent.history
        if len(self.history) >= 6 and all([oh[i] != oh[i+1] for i in range(len(oh)-1)]):
            return 'D'
        return 'C'


class MathConstantHunter(Player):
    """A player who hunts for mathemtical constant players."""

    name = "Math Constant Hunter"
    memory_depth = float('inf')  # Long memory

    # We need to make sure this is not marked as stochastic.
    def __init__(self):
        Player.__init__(self)
        self.stochastic = False

    def strategy(self, opponent):
        """
        Check whether the number of cooperations in the first and second halves
        of the history are close. The variance of the uniform distribution (1/4)
        is a reasonable delta but use something lower for certainty and avoiding
        false positives. Rhis approach will also detect a lot of random players.
        """

        n = len(self.history)
        if n >= 8 and opponent.cooperations and opponent.defections:

            start1, end1 = 0, n // 2
            start2, end2 = n // 4, 3 * n // 4
            start3, end3 = n // 2, n
            count1 = opponent.history[start1: end1].count('C') + self.history[start1: end1].count('C')
            count2 = opponent.history[start2: end2].count('C') + self.history[start2: end2].count('C')
            count3 = opponent.history[start3: end3].count('C') + self.history[start3: end3].count('C')
            ratio1 = 0.5 * count1 / (end1 - start1)
            ratio2 = 0.5 * count2 / (end2 - start2)
            ratio3 = 0.5 * count3 / (end3 - start3)
            if abs(ratio1 - ratio2) < 0.2 and abs(ratio1 - ratio3) < 0.2:
                return 'D'

        return 'C'


class RandomHunter(Player):
    """A player who hunts for random players."""

    name = "Random Hunter"
    memory_depth = float('inf')  # Long memory

    # We need to make sure this is not marked as stochastic.
    def __init__(self):
        Player.__init__(self)
        self.stochastic = False

    def strategy(self, opponent):
        """
        A random player is unpredictable, which means the conditional frequency
        of cooperation after cooperation, and defection after defections, should
        be close to 50%... although how close is debatable.
        """

        n = len(self.history)
        if n > 10:
            probabilities = []
            if self.history[:-1].count('C') > 5:
                countCC = len([i for i in range(n-1) if self.history[i] == "C" and opponent.history[i+1] == "C"])
                probabilities.append(1.0 * countCC / self.history[:-1].count("C"))
            if self.history[:-1].count('D') > 5:
                countDD = len([i for i in range(n-1) if self.history[i] == "D" and opponent.history[i+1] == "D"])
                probabilities.append(1.0 * countDD / self.history[:-1].count("D"))

            if probabilities and all([abs(p - 0.5) < 0.25 for p in probabilities]):
                return 'D'

        return 'C'


class MetaHunter(MetaPlayer):
    """A player who uses a selection of hunters."""

    name = "Meta Hunter"
    memory_depth = float('inf')  # Long memory

    def __init__(self):

        # We need to make sure this is not marked as stochastic.
        self.stochastic = False

        # Notice that we don't include the cooperator hunter, because it leads to excessive
        # defection and therefore bad performance against unforgiving strategies. We will stick
        # to hunters that use defections as cues. However, a really tangible benefit comes from
        # combining Random Hunter and Math Constant Hunter, since together they catch strategies
        # that are lightly randomized but still quite constant (the tricky/suspecious ones).
        self.team = [DefectorHunter, AlternatorHunter, RandomHunter, MathConstantHunter]

        MetaPlayer.__init__(self)

    @staticmethod
    def meta_strategy(results, opponent):

        # If any of the hunters smells prey, then defect!
        if 'D' in results:
            return 'D'

        # Tit-for-tat might seem like a better default choice, but in many cases it complicates
        # the heuristics of hunting and creates fale-positives. So go ahead and use it, but only
        # for longer histories.
        if len(opponent.history) > 100:
            return 'D' if opponent.history[-1:] == ['D'] else 'C'
        else:
            return 'C'



from __future__ import division
import copy


class RoundRobin(object):
    """A class to define play a round robin game of players"""

    def __init__(self, players, game, turns, deterministic_cache=None,
                 cache_mutable=True, noise=0):
        """Initialise the players, game and deterministic cache"""
        self.players = players
        self.nplayers = len(players)
        self.game = game
        self.turns = turns
        if deterministic_cache is None:
            self.deterministic_cache = {}
        else:
            self.deterministic_cache = deterministic_cache
        self.cache_mutable = cache_mutable
        self._noise = noise

    def play(self):
        """Plays a round robin where each match lasts turns.

        We cache scores for pairs of deterministic strategies, since the
        outcome will always be the same.

        Notice also that we need to handle self-interactions with some special
        exceptions due to the way gameplay is coded within Player.

        Returns the total payoff matrix.
        """

        payoff = self._empty_matrix(self.nplayers, self.nplayers)
        cooperation = self._empty_matrix(self.nplayers, self.nplayers)

        for player1_index in range(self.nplayers):
            for player2_index in range(player1_index, self.nplayers):
                scores, cooperation_rates = self._score_single_interaction(
                    player1_index, player2_index)
                self._update_matrices(
                    player1_index, player2_index, scores, payoff,
                    cooperation_rates, cooperation)

        return {'payoff': payoff, 'cooperation': cooperation}

    def _empty_matrix(self, rows, columns):
        return [[0 for j in range(columns)] for i in range(rows)]

    def _score_single_interaction(self, player1_index, player2_index):
        player1, player2, classes = self._pair_of_players(
            player1_index, player2_index)
        play_required = (
            self._stochastic_interaction(player1, player2) or
            classes not in self.deterministic_cache)
        if play_required:
            scores, cooperation_rates = self._play_single_interaction(
                player1, player2, classes)
        else:
            scores = self.deterministic_cache[classes]['scores']
            cooperation_rates = (
                self.deterministic_cache[classes]['cooperation_rates'])
        return scores, cooperation_rates

    def _update_matrices(self, player1_index, player2_index, scores,
                         payoffs, cooperation_rates, cooperation):
        # For self-interactions we can take the average of the two
        # sides, which should improve the averaging a bit.
        if not self._noise and player1_index == player2_index:
            payoffs[player1_index][player2_index] = (
                0.5 * (scores[0] + scores[1]))
        else:
            payoffs[player1_index][player2_index] = scores[0]
            payoffs[player2_index][player1_index] = scores[1]

        cooperation[player1_index][player2_index] = cooperation_rates[0]
        cooperation[player2_index][player1_index] = cooperation_rates[1]

    def _pair_of_players(self, player1_index, player2_index):
        player1 = self.players[player1_index]
        class1 = player1.__class__
        if player1_index == player2_index:
            player2 = copy.deepcopy(player1)
            class2 = class1
        else:
            player2 = self.players[player2_index]
            class2 = player2.__class__
        return player1, player2, (class1, class2)

    def _stochastic_interaction(self, player1, player2):
        return (
            self._noise or
            player1.stochastic or
            player2.stochastic)

    def _play_single_interaction(self, player1, player2, classes):
        turn = 0
        player1.reset()
        player2.reset()
        while turn < self.turns:
            turn += 1
            player1.play(player2, self._noise)
        scores = self._calculate_scores(player1, player2)
        cooperation_rates = (
            self._calculate_cooperation(player1),
            self._calculate_cooperation(player2))
        if self._cache_update_required(player1, player2):
            self.deterministic_cache[classes] = {
                'scores': scores,
                'cooperation_rates': cooperation_rates}
        return scores, cooperation_rates

    def _calculate_scores(self, p1, p2):
        """Calculates the score for two players based their history"""
        s1, s2 = 0, 0
        for pair in zip(p1.history, p2.history):
            score = self.game.score(pair)
            s1 += score[0]
            s2 += score[1]
        return s1, s2

    def _cache_update_required(self, p1, p2):
        return (
            not self._noise and
            self.cache_mutable and
            not (p1.stochastic or p2.stochastic))

    def _calculate_cooperation(self, player):
        return player.history.count('C')



import inspect
import random

C, D = 'C', 'D'
flip_dict = {C: D, D: C}


def update_histories(player1, player2, move1, move2):
    """Updates histories and cooperation / defections counts following play."""
    # Update histories
    player1.history.append(move1)
    player2.history.append(move2)
    # Update player counts of cooperation and defection
    if move1 == C:
        player1.cooperations += 1
    elif move1 == D:
        player1.defections += 1
    if move2 == C:
        player2.cooperations += 1
    elif move2 == D:
        player2.defections += 1


class Player(object):
    """A class for a player in the tournament.

    This is an abstract base class, not intended to be used directly.
    """

    name = "Player"

    def __init__(self):
        """Initiates an empty history and 0 score for a player."""
        self.history = []
        self.stochastic = "random" in inspect.getsource(self.__class__)
        self.tournament_length = -1
        if self.name == "Player":
            self.stochastic = False
        self.cooperations = 0
        self.defections = 0

    def __repr__(self):
        """The string method for the strategy."""
        return self.name

    def _add_noise(self, noise, s1, s2):
        r = random.random()
        if r < noise:
            s1 = flip_dict[s1]
        r = random.random()
        if r < noise:
            s2 = flip_dict[s2]
        return s1, s2

    def strategy(self, opponent):
        """This is a placeholder strategy."""
        return None

    def play(self, opponent, noise=0):
        """This pits two players against each other."""
        s1, s2 = self.strategy(opponent), opponent.strategy(self)
        if noise:
            s1, s2 = self._add_noise(noise, s1, s2)
        update_histories(self, opponent, s1, s2)

    def reset(self):
        """Resets history.

        When creating strategies that create new attributes then this method should be
        re-written (in the inherited class) and should not only reset history but also
        rest all other attributes.
        """
        self.history = []
        self.cooperations = 0
        self.defections = 0



from __future__ import absolute_import

from .random_ import random_choice

from .plot import *
from .game import *
from .player import *
from .mock_player import *
from .round_robin import *
from .tournament import *
from .tournament_manager import *
from .strategies import *
from .ecosystem import *


from distutils.core import setup

setup(
    name='Axelrod',
    version='0.0.10',
    author='Vince Knight, Owen Campbell, Karol Langner',
    author_email=('axelrod-python@googlegroups.com'),
    packages=['axelrod', 'axelrod.strategies', 'axelrod.tests'],
    scripts=['run_axelrod'],
    url='http://axelrod.readthedocs.org/',
    license='The MIT License (MIT)',
    description='Reproduce the Axelrod iterated prisoners dilemma tournament',
    install_requires=[
        "matplotlib >= 1.4.2",
    ],
)



from axelrod import Player


class Appeaser(Player):
    """A player who tries to guess what the opponent wants.

    Switch the behaviour every time the opponent plays 'D'.
    Start with 'C', switch between 'C' and 'D' when opponent plays 'D'.
    """

    name = 'Appeaser'
    memory_depth = float('inf')  # Depends on internal memory.

    def strategy(self, opponent):
        if not len(self.history):
            self.move = 'C'
        else:
            if opponent.history[-1] == 'D':
                if self.move == 'C':
                    self.move = 'D'
                else:
                    self.move = 'C'
        return self.move
